"""create local hpo experiment table

Revision ID: c22620b97e39
Revises: 6dfd859f129e
Create Date: 2021-05-18 17:32:53.814373

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c22620b97e39'
down_revision = '6dfd859f129e'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('local_hpo_experiment',
    sa.Column('id', sa.String(length=8), nullable=False),
    sa.Column('port', sa.Integer(), nullable=False),
    sa.Column('mode', sa.String(length=8), nullable=False),
    sa.Column('status', sa.String(length=16), nullable=False),
    sa.Column('project_id', sa.String(length=32), nullable=False),
    sa.Column('project_name', sa.String(length=255), nullable=False),
    sa.Column('algorithm_id', sa.String(length=32), nullable=False),
    sa.Column('algorithm_dir', sa.String(length=4096), nullable=True),
    sa.Column('algorithm_archive', sa.String(length=4096), nullable=True),
    sa.Column('dataset_id', sa.String(length=32), nullable=False),
    sa.Column('dataset_dir', sa.String(length=4096), nullable=True),
    sa.Column('dataset_archive', sa.String(length=4096), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('local_hpo_experiment')
    # ### end Alembic commands ###



from requests.exceptions import RequestException
import responses

from anylearn.interfaces.user import User
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from tests.base_test_case import BaseTestCase


class TestUser(BaseTestCase):
    @responses.activate
    def test_create_user_ok(self):
        responses.add(responses.POST, url=self._url("user/registry"),
                      json={'data': "USER123", 'message': "用户添加成功"},
                      status=200)
        user = User(username="test_username",
                    password="123456", email="123@abc.com")
        res = user.save()
        self.assertTrue(res, True)

    def test_create_user_empty_username_password_email(self):
        user = User()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            user.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "User缺少必要字段：['username', 'password', 'email']")

    @responses.activate
    def test_create_user_unknown_response(self):
        responses.add(responses.POST, url=self._url("user/registry"),
                      json={'msg': "Unknown response"}, status=201)
        user = User(username="test_username",
                    password="123456", email="abc@123.com")
        with self.assertRaises(AnyLearnException) as ctx:
            user.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_user_detail_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("user?id=USER123"),
                      match_querystring=True,
                      json=[
                          {
                              "collected_algorithm": "[]",
                              "email": "123@abc.com",
                              "id": "USER123",
                              "namespace": "xln111",
                              "own_algorithms": [
                                  "ALGO1",
                                  "ALGO2",
                                  "ALGO3"
                              ],
                              "own_datasets": [
                                  "DSET1",
                                  "DSET2",
                                  "DSET3"
                              ],
                              "own_files": [
                                  "FILE1",
                                  "FILE2",
                                  "FILE3"
                              ],
                              "own_models": [
                                  "MODE1",
                                  "MODE2",
                                  "MODE3"
                              ],
                              "role": "admin,user",
                              "username": "xlearn"
                          }
                        ],
                    status=200)
        user = User(id="USER123")
        user.get_detail()
        self.assertIsInstance(user.own_datasets, list)
        if isinstance(user.own_datasets, list):
            self.assertEqual(len(user.own_datasets), 3)
        self.assertEqual(user.own_datasets[0], "DSET1")
        self.assertEqual(user.id, "USER123")
        self.assertEqual(user.role, "admin,user")
        self.assertEqual(user.email, "123@abc.com")
        self.assertEqual(user.username, "xlearn")

    @responses.activate
    def test_get_user_detail_no_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("user?id=USER123"),
                      match_querystring=True,
                      json=[
                          {
                              "collected_algorithm": "[]",
                              "email": "123@abc.com",
                              "id": "USER123",
                              "namespace": "xln111",
                              "own_algorithms": [
                                  "ALGO1",
                                  "ALGO2",
                                  "ALGO3"
                              ],
                              "own_datasets": [
                                  "DSET1",
                                  "DSET2",
                                  "DSET3"
                              ],
                              "own_files": [
                                  "FILE1",
                                  "FILE2",
                                  "FILE3"
                              ],
                              "own_models": [
                                  "MODE1",
                                  "MODE2",
                                  "MODE3"
                              ],
                              "role": "admin,user",
                              "username": "xlearn"
                          }
                        ],
                      status=200)
        user = User(id="USER123", load_detail=True)
        self.assertIsInstance(user.own_datasets, list)
        if isinstance(user.own_datasets, list):
            self.assertEqual(len(user.own_datasets), 3)
        self.assertEqual(user.id, "USER123")
        self.assertEqual(user.role, "admin,user")
        self.assertEqual(user.email, "123@abc.com")
        self.assertEqual(user.username, "xlearn")

    @responses.activate
    def test_get_user_detail_ko_403(self):
        responses.add(responses.GET, url=self._url("user?id=USER403"),
                      match_querystring=True, status=403)
        user = User(id="USER403")
        with self.assertRaises(RequestException) as ctx:
            user.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, RequestException)
        self.assertEqual(e.response.status_code, 403)

    @responses.activate
    def test_get_user_detail_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("user?id=USER250"),
                      match_querystring=True, json={'msg': "Unknown response"},
                      status=200)
        user = User(id="USER250")
        with self.assertRaises(AnyLearnException) as ctx:
            user.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_user_detail_empty_list(self):
        responses.add(responses.GET, url=self._url("user?id=USER250"),
                      match_querystring=True, json=[],
                      status=200)
        user = User(id="USER250")
        with self.assertRaises(AnyLearnException) as ctx:
            user.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_change_password_ok(self):
        responses.add(responses.PUT, url=self._url("user/password"),
                      json={'data': "USER123", 'message': "密码修改成功"},
                      status=200)
        user = User(id="USER123")
        res = user.change_password("old_password", "new_password")
        self.assertEqual(res['data'], "USER123")

    def test_change_password_empty_id(self):
        user = User()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            user.change_password("old_password", "new_password")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "User缺少必要字段：['id']")

    @responses.activate
    def test_change_password_unknown_response(self):
        responses.add(responses.PUT, url=self._url("user/password"),
                      json={'msg': "Unknown response"}, status=201)
        user = User(id="USER123")
        with self.assertRaises(AnyLearnException) as ctx:
            user.change_password("old_password", "new_password")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_list_user_ok(self):
        responses.add(responses.GET, url=self._url("user/list"),
                      json=[
                          {
                              'id': "USER001",
                              'username': "TestUser1",
                              'email': "123@abc.com",
                              'role': "user",
                          },
                          {
                              'id': "USER002",
                              'username': "TestUser2",
                              'email': "456@abc.com",
                              'role': "user2",
                          },
                        ],
                      status=200)
        users = User.get_list()
        self.assertIsInstance(users, list)
        self.assertEqual(len(users), 2)
        self.assertIsInstance(users[0], User)
        self.assertIsInstance(users[1], User)
        self.assertEqual(users[0].id, "USER001")
        self.assertEqual(users[1].id, "USER002")

    @responses.activate
    def test_list_user_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("user/list"),
                      json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            User.get_list()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_user_collection_ok(self):
        responses.add(responses.GET, url=self._url("user/collection?collection=USER001,USER002"),
                      match_querystring=True,
                      json=[
                          {
                              'id': "USER001",
                              'role': "user1",
                              'email': "123@abc.com",
                              'username': "xlearn1"
                          },
                          {
                              'id': "USER002",
                              'role': "user2",
                              'email': "456@abc.com",
                              'username': "xlearn2"
                          }
                        ],
                      status=200)
        users = User.user_collection("USER001,USER002")
        self.assertIsInstance(users, list)
        self.assertEqual(len(users), 2)
        self.assertIsInstance(users[0], User)
        self.assertIsInstance(users[1], User)
        self.assertEqual(users[0].id, "USER001")
        self.assertEqual(users[1].id, "USER002")

    @responses.activate
    def test_user_collection_ko_403(self):
        responses.add(responses.GET, url=self._url("user/collection?collection=USER403"),
                      match_querystring=True, status=403)
        with self.assertRaises(RequestException) as ctx:
            User.user_collection("USER403")
        e = ctx.exception
        self.assertIsInstance(e, RequestException)
        self.assertEqual(e.response.status_code, 403)

    @responses.activate
    def test_user_collection_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("user/collection?collection=USER250"),
                      match_querystring=True, json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            User.user_collection("USER250")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_user_update_ok(self):
        responses.add(responses.PUT, url=self._url("user"),
                      json={'data': "USER123", 'message': "用户信息修改成功"},
                      status=200)
        user = User(id="USER123", username="test",
                    email="123@abc.com", role="user")
        res = user.save()
        self.assertTrue(res)

    @responses.activate
    def test_user_update_unknown_response(self):
        responses.add(responses.PUT, url=self._url("user"),
                      json={'msg': "Unknown response"}, status=201)
        user = User(id="USER201", username="test",
                    email="123@abc.com", role="user")
        with self.assertRaises(AnyLearnException) as ctx:
            user.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_user_delete_ok(self):
        responses.add(responses.DELETE, url=self._url("user"),
                      json={'data': "USER123", 'message': "用户删除成功"},
                      status=200)
        user = User(id="USER123")
        res = user.delete()
        self.assertTrue(res)

    def test_user_delete_empty_id(self):
        user = User()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            user.delete()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "User缺少必要字段：['id']")

    @responses.activate
    def test_user_delete_unknown_response(self):
        responses.add(responses.DELETE, url=self._url("user"),
                      json={'msg': "Unknown response"}, status=201)
        user = User(id="USER201")
        with self.assertRaises(AnyLearnException) as ctx:
            user.delete()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_user_check_ok(self):
        responses.add(responses.GET, url=self._url("user/check"),
                      json=True,
                      status=200)
        res = User.user_check(username="test")
        self.assertIsInstance(res,bool)
        self.assertEqual(res, True)

    def test_user_check_empty_username(self):
        with self.assertRaises(AnyLearnException) as ctx:
            User.user_check(username="")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "username值不能为空")

    @responses.activate
    def test_user_check_unknown_response(self):
        responses.add(responses.GET, url=self._url("user/check"),
                      json={'msg': "Unknown response"}, status=201)
        with self.assertRaises(AnyLearnException) as ctx:
            User.user_check(username="username201")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")



from .algorithm import Algorithm
from .dataset import Dataset
from .file import File
from .model import Model
from .resource_uploader import ResourceUploader, AsyncResourceUploader, SyncResourceUploader
from .resource_downloader import ResourceDownloader, AsyncResourceDownloader, SyncResourceDownloader
from .resource import Resource, ResourceState



import abc
import os
import time
import zipfile
from pathlib import Path
from typing import Optional, Union

import anylearn.env as env
from anylearn.config import AnylearnConfig
from anylearn.sdk.artifacts.compression import Compression
from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.cmd import cmd_error, cmd_info
from anylearn.utils.errors import (
    AnyLearnException,
    AnyLearnNotSupportedException,
)
from anylearn.utils.util import ensure_dir


class Artifact(abc.ABC):
    """
    Attributes
    ----------
    id
        资源的唯一标识符，自动生成，由类名前四位+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
    name
        资源的名称（长度1~50）
    description
        资源描述（长度最大255）
    state
        资源状态
    public
        数据集是否为公开（默认为False）
    upload_time
        资源上传时间
    filename
        下一步中会被分片上传的资源的完整名（包括扩展名）
    is_zipfile
        是否为zip文件
    file_path
        资源文件路径
    size
        资源文件大小
    creator_username
        创建者用户名
    """

    """具体资源信息配置"""
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'filename'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'public', 'owner',
                       'filename'],
            'update': ['id', 'name', 'description', 'public', 'owner'],
        },
    }

    def __init__(self, *_, **kwargs):
        """
        Parameters
        ----------
        load_detail
            初始化时是否加载详情
        """
        self.id = kwargs.pop('id', None)
        self.name = kwargs.pop('name', None)
        self.description = kwargs.pop('description', None)
        self.state = kwargs.pop('state', None)
        self.public = kwargs.pop('public', False)
        self.uploaded_at = kwargs.pop('upload_time', None)
        self.size = kwargs.pop('size', None)
        self.creator_username = kwargs.pop('creator_username', None)
        self.__original_filename = kwargs.pop('filename', None)
        self.__original_is_zipfile = kwargs.pop('is_zipfile', None)

    def ls(self, sub_path: str=None):
        """
        查询文件列表

        Parameters
        ----------
        sub_path : :obj:`str`
            资产文件目录内的子路径（相对路径），默认为空，即资产根目录
        """
        if not self.id:
            cmd_error("Artifact ID is required")
        return get_with_token(
            f"{url_base()}/resource/listdir",
            params={
                'file_id': self.id,
                'subpath': sub_path,
            },
        )

    def upload(self, path: Union[os.PathLike, str]) -> bool:
        raise AnyLearnNotSupportedException

    def download(
        self,
        dest_dir: Optional[Union[os.PathLike, str]]=None,
        sub_path: Optional[str]=None,
    ) -> str:
        if not self.id:
            cmd_error("Artifact ID is required")
        local_path = self._get_local_path_from_env()
        if not local_path:
            archive_path = self.compress_and_download(sub_path=sub_path)
            root_dir = dest_dir or (
                AnylearnConfig.workspace_path / "artifacts"
            )
            ensure_dir(root_dir)
            with zipfile.ZipFile(archive_path, "r") as zip_handle:
                zip_handle.extractall(root_dir)
            local_path = str(root_dir / self.id)
        return local_path

    def _get_local_path_from_env(self) -> Optional[Path]:
        paths = env.get_artifact_paths()
        return paths.get(self.id, None)

    def compress_and_download(
        self,
        dest_dir: Optional[Union[os.PathLike, str]]=None,
        sub_path: Optional[str]=None,
        polling: Union[float, int]=5,
    ) -> Path:
        cmd_info("Compressing...")
        t_elapsed = 0
        compression = Compression.create(self.id, sub_path)
        while not compression.ready_to_download():
            time.sleep(polling)
            t_elapsed += polling
            compression.reload()
            if compression.on_error():
                raise AnyLearnException('Compression failed')
            elif t_elapsed % 10 == 0:
                cmd_info(
                    "| Still waiting compression to be ready "
                    f"({t_elapsed}s)"
                )

        cmd_info("Downloading...")
        return compression.download(dest_dir)

    def __repr__(self) -> str:
        d = self.__dict__
        kv = ", ".join([f"{k}={d[k]!r}" for k in d])
        return f"{self.__class__.__name__}({kv})"

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__}"
            f" fullname={self.creator_username}/{self.name}"
            ">"
        )


class ArtifactState:
    """
    资产状态标识：

    - 1(CREATED)表示已创建
    - 2(UPLOADING)表示上传中
    - 3(READY)表示就绪
    - -1(DELETED)表示已删除
    - -2(ERROR)表示出错
    """
    CREATED = 1
    UPLOADING = 2
    READY = 3
    DELETED = -1
    ERROR = -2



from __future__ import annotations

from anylearn.sdk.artifacts.artifact import Artifact
from anylearn.utils.api import get_with_token, url_base
from anylearn.utils.errors import AnyLearnException


class FileArtifact(Artifact):
    @classmethod
    def from_id(cls, id_: str) -> FileArtifact:
        res = get_with_token(
            f"{url_base()}/file/query",
            params={'id': id_},
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        return FileArtifact(**res[0])



import os

from anylearn.config import AnylearnConfig, init_sdk_incontainer
from anylearn.interfaces import TrainTask
from anylearn.utils import logger
from anylearn.utils.errors import AnyLearnException


INCONTAINER_TRAIN_TASK_ID = os.environ.get('task_id', None)
INCONTAINER_TRAIN_TASK_SECRET = os.environ.get('secret_key', None)


def report_intermediate_metric(metric: float):
    """
    向Anylearn后端引擎汇报模型训练的中间结果指标。

    Parameters
    ----------
    metric:
        中间结果指标的值，浮点数类型。
    """
    __init()
    train_task = get_incontainer_train_task()
    if train_task and train_task.id:
        train_task.report_intermediate_metric(metric)
    else:
        logger.info(f"Reported intermediate metric: {metric}")


def report_final_metric(metric: float):
    """
    向Anylearn后端引擎汇报模型训练的最终结果指标。

    Parameters
    ----------
    metric:
        最终结果指标的值，浮点数类型。
    """
    __init()
    train_task = get_incontainer_train_task()
    if train_task and train_task.id:
        train_task.report_final_metric(metric)
    else:
        logger.info(f"Reported final metric: {metric}")


def get_incontainer_train_task():
    """
    从Anylearn后端引擎的当前训练环境（容器）中获取训练任务对象。

    Returns
    -------
    TrainTask
        当前训练环境中的训练对象。
    """
    if not all([
        INCONTAINER_TRAIN_TASK_ID,
        INCONTAINER_TRAIN_TASK_SECRET,
    ]):
        raise AnyLearnException("无法获取训练环境")
    return TrainTask(id=INCONTAINER_TRAIN_TASK_ID,
                     secret_key=INCONTAINER_TRAIN_TASK_SECRET)


def __init():
    if AnylearnConfig.cluster_address:
        return
    namespace = os.environ.get('BACKEND_NAMESPACE', "STANDALONE")
    init_sdk_incontainer(f"http://anylearn-backend.{namespace}")



import ast
import fnmatch
import os
import re
import sys
from pathlib import Path
from typing import Any, Dict, List, Set

from anylearn.utils.errors import AnyLearnException


class AnylearnImportVisitor(ast.NodeVisitor):

    def __init__(self) -> None:
        super().__init__()
        self.visited: bool = False
        self.imported_anylearn_as: Set[str] = set()
        self.imported_elements_as: Dict[str, Set[str]] = {}

    def visit(self, node: ast.AST) -> Any:
        self.visited = True
        return super().visit(node)

    def visit_Import(self, node: ast.Import):
        for child in node.names:
            if child.name:
                first, middle, rest = child.name.partition(".")
                if first == "anylearn" and not middle and not rest:
                    self.imported_anylearn_as.add(child.asname or child.name)
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module and node.level == 0:
            first, _, _ = node.module.partition(".")
            if first == "anylearn":
                for child in node.names:
                    k = f"{node.module}.{child.name}"
                    if not k in self.imported_elements_as:
                        self.imported_elements_as[k] = set()
                    self.imported_elements_as[k].add(child.asname or child.name)
        self.generic_visit(node)


class ArtifactUsageVisitor(ast.NodeVisitor):

    def __init__(
        self,
        import_visitor: AnylearnImportVisitor,
    ) -> None:
        super().__init__()
        self.import_visitor: AnylearnImportVisitor = import_visitor
        self.parsed_import_calls: bool = False
        self.calls: Dict[str, Dict[str, Any]] = {
            'datasets': {
                'attribute_call': "get_dataset",
                'original_name_calls': [
                    "anylearn.get_dataset",
                ],
                'imported_name_calls': [],
                'argn': 0,
                'kwarg': 'full_name',
            },
            'models': {
                'attribute_call': "get_model",
                'original_name_calls': [
                    "anylearn.get_model",
                ],
                'imported_name_calls': [],
                'argn': 0,
                'kwarg': 'full_name',
            },
            'task_outputs': {
                'attribute_call': "get_task_output",
                'original_name_calls': [
                    "anylearn.get_task_output",
                    "anylearn.sdk.get_task_output",
                ],
                'imported_name_calls': [],
                'argn': 0,
                'kwarg': 'task_id',
            }
        }
        self.fullnames: Dict[str, Set[str]] = {
            'datasets': set(),
            'models': set(),
            'task_outputs': set(),
        }
        self.unfetchable: List[Any] = []

    def visit(self, node: ast.AST) -> Any:
        if not self.import_visitor.visited:
            raise AnyLearnException(
                "AnylearnImportVisitor should visit AST first"
            )
        if not self.parsed_import_calls:
            for _t, calls in self.calls.items():
                for call in calls['original_name_calls']:
                    _i = self.import_visitor.imported_elements_as.get(call, [])
                    self.calls[_t]['imported_name_calls'].extend(_i)
            self.parsed_import_calls = True
        return super().visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        for artifact_type in self.calls.keys():
            self.__fetch_full_name_from_call__(
                call=node,
                artifact_type=artifact_type,
            )
        self.generic_visit(node)

    def __fetch_full_name_from_call__(self, call: ast.Call, artifact_type: str):
        f = call.func

        # Calls like foo.bar()
        # where "bar" is raw string in ast.Attribute.attr
        # and foo is in ast.Name.id in ast.Attribute.value
        valid_attr_call = isinstance(f, ast.Attribute) and \
            isinstance(f.value, ast.Name) and \
            f.value.id in self.import_visitor.imported_anylearn_as and \
            f.attr == self.calls[artifact_type]['attribute_call']

        # Calls like print()
        valid_name_call = isinstance(f, ast.Name) and \
            f.id in self.calls[artifact_type]['imported_name_calls']

        if valid_attr_call or valid_name_call:
            _N = self.calls[artifact_type]['argn']
            _KW = self.calls[artifact_type]['kwarg']
            if len(call.args) > _N:
                _fn = self.__fetch_full_name_from_arg__(call.args[_N])
            else:
                kw = next((kw for kw in call.keywords if kw.arg == _KW), None)
                _fn = self.__fetch_full_name_from_kwarg__(kw)
            if _fn:
                self.fullnames[artifact_type].add(_fn)

    def __fetch_full_name_from_arg__(self, arg: ast.AST):
        if not arg:
            return None
        if sys.version_info < (3, 8):
            if isinstance(arg, ast.Str):
                return arg.s
            else:
                self.unfetchable.append(arg)
                return None
        else:
            if isinstance(arg, ast.Constant):
                return arg.value
            else:
                self.unfetchable.append(arg)
                return None

    def __fetch_full_name_from_kwarg__(self, kw: ast.keyword):
        if not kw:
            return None
        if sys.version_info < (3, 8):
            if isinstance(kw.value, ast.Str):
                return kw.value.s
            else:
                self.unfetchable.append(kw)
                return None
        else:
            if isinstance(kw.value, ast.Constant):
                return kw.value.value
            elif not isinstance(kw.value, ast.Constant):
                self.unfetchable.append(kw)
                return None


IGNORE = [
    "__pycache__",
    "build",
    "develop-eggs",
    "dist",
    "eggs",
    ".eggs",
    "sdist",
    "wheels",
    "*.egg-info",
    "htmlcov",
    ".tox",
    ".nox",
    "cover",
    "_build",
    "__pypackages__",
    ".env",
    ".venv",
    "env",
    "venv",
    "ENV",
    ".idea",
    ".vscode",
]


class ArtifactCollector:

    def __init__(self) -> None:
        self.unfetchable: Dict[str, ast.AST] = {}

    def collect(
        self,
        path: os.PathLike,
        encoding: str='utf8',
        follow_links: bool=False,
    ) -> Dict[str, Set[str]]:
        path = Path(path)
        if path.is_file():
            return self.__collect_in_file__(
                file_path=path,
                encoding=encoding,
            )
        elif path.is_dir():
            return self.__collect_in_dir__(
                dir_path=path,
                encoding=encoding,
                follow_links=follow_links,
            )
        else:
            raise NotADirectoryError

    def __collect_in_file__(
        self,
        file_path: os.PathLike,
        encoding: str='utf8',
    ) -> Dict[str, Set[str]]:
        import_visitor = AnylearnImportVisitor()
        artifact_visitor = ArtifactUsageVisitor(import_visitor)
        _fp = Path(file_path)
        with _fp.open("r", encoding=encoding) as f:
            contents = f.read()
            astroot = ast.parse(contents)
        import_visitor.visit(astroot)
        artifact_visitor.visit(astroot)
        self.unfetchable[_fp.absolute()] = artifact_visitor.unfetchable.copy()
        return artifact_visitor.fullnames

    def __collect_in_dir__(
        self,
        dir_path: os.PathLike,
        encoding: str='utf8',
        follow_links: bool=False,
    ) -> Dict[str, Set[str]]:
        fullnames = {
            'datasets': set(),
            'models': set(),
            'task_outputs': set(),
        }

        ignore = IGNORE
        ignore_pattern = r"|".join([fnmatch.translate(i) for i in ignore])

        walk = os.walk(dir_path, followlinks=follow_links, topdown=True)
        for root, dirs, files in walk:
            # In-place update `dirs` to exclude ignored dirs
            dirs[:] = [d for d in dirs if not re.match(ignore_pattern, d)]
            files = [f for f in files if Path(f).suffix == ".py"]

            for file_name in files:
                file = Path(root) / file_name
                _fn = self.__collect_in_file__(
                    file_path=file,
                    encoding=encoding,
                )
                fullnames['datasets'].update(_fn['datasets'])
                fullnames['models'].update(_fn['models'])
                fullnames['task_outputs'].update(_fn['task_outputs'])
        return fullnames
